from typing import Optional
import os
import shutil
from ._load_http_file import _load_http_file
from ._load_github_file import _load_github_file
from .load_file_local import load_file_local
from ._load_sha1_file_from_cloud import _load_sha1_file_from_cloud


def load_file(
    uri: str,
    *,
    dest: Optional[str] = None,
    local_only: bool = False,
    _get_info: bool = False,
    verbose: bool = False,
    zone: Optional[str] = None,
) -> Optional[str]:
    if uri.startswith("gh://"):
        if _get_info:
            raise Exception("Cannot use _get_info for this uri")
        return _load_github_file(uri)

    elif uri.startswith("http://") or uri.startswith("https://"):
        if _get_info:
            raise Exception("Cannot use _get_info for this uri")
        return _load_http_file(uri)

    elif local_only:
        if _get_info:
            raise Exception("Cannot use _get_info with local_only")
        return load_file_local(uri, dest=dest)
    elif uri.startswith("/"):
        if _get_info:
            raise Exception("Cannot use _get_info for this uri")
        if os.path.exists(uri):
            if dest is not None:
                shutil.copyfile(uri, dest)
                return dest
            return uri
        else:
            return None
    elif uri.startswith("sha1://"):
        if not _get_info:
            x = load_file_local(uri, dest=dest)
            if x is not None:
                return x
        sha1 = uri.split("?")[0].split("/")[2]
        fname_or_info = _load_sha1_file_from_cloud(
            sha1, verbose=verbose, dest=dest, _get_info=_get_info, zone=zone
        )
        if fname_or_info is None:
            return None
        elif isinstance(fname_or_info, dict):
            if not _get_info:
                raise Exception("Unexpected dict in load_file")
            return fname_or_info  # type: ignore
        elif isinstance(fname_or_info, str):
            if _get_info:
                raise Exception("Unexpected string in load_file")
            return fname_or_info
        else:
            raise Exception("Unexpected return value from _load_sha1_file_from_cloud")
    elif uri.startswith("kachery:"):
        parts = uri.split(":")
        if len(parts) < 4:
            raise Exception(f"Unexpected uri: {uri}")
        zone = parts[1]
        alg = parts[2]
        hash0 = parts[3]
        # label = parts[4] if len(parts) > 4 else None
        uri2 = f"{alg}://{hash0}"
        return load_file(
            uri2,
            dest=dest,
            local_only=local_only,
            _get_info=_get_info,
            verbose=verbose,
            zone=zone,
        )
    else:
        raise Exception(f"Unexpected uri: {uri}")


def load_file_info(uri: str) -> dict:
    return load_file(uri, _get_info=True)  # type: ignore



import random


def _random_string(num_chars: int) -> str:
    chars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
    return "".join(random.choice(chars) for _ in range(num_chars))



from tempfile import TemporaryDirectory
import requests
from .load_file_local import load_file_local


def _load_http_file(url: str):
    from .store_file_local import store_file_local

    with TemporaryDirectory(prefix="load_http_file") as tmpdir:
        tmp_filename = f"{tmpdir}/file.dat"
        with requests.get(url, stream=True) as r:
            if r.status_code == 404:
                raise Exception(f"File not found: {url}")
            r.raise_for_status()
            with open(tmp_filename, "wb") as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)
        uri = store_file_local(tmp_filename)
        return load_file_local(uri)



from typing import Union
from urllib.parse import quote


def _form_uri(
    *, alg: str, hash0: str, label: Union[str, None], uri_type: int, zone: str
) -> str:
    if uri_type == 1:
        uri = f"{alg}://{hash0}"
        if label is not None:
            uri = f"{uri}?label={quote(label)}"
        return uri
    elif uri_type == 2:
        uri = f"kachery:{zone}:{alg}:{hash0}"
        if label is not None:
            uri = f"{uri}:{quote(label)}"
        return uri
    else:
        raise Exception(f"Unexpected uri_type: {uri_type}")



import os

from .client.load_file import load_file, load_file_info  # noqa: F401
from .client.cat_file import cat_file  # noqa: F401
from .client.store_file_local import store_file_local  # noqa: F401
from .client.store_file import store_file  # noqa: F401
from .client.core import store_text, store_json, store_npy  # noqa: F401
from .client.core import load_text, load_json, load_npy  # noqa: F401
from .client.core import (
    store_text_local,
    store_json_local,
    store_npy_local,
)  # noqa: F401
from .client.load_bytes import load_bytes  # noqa: F401
from .client.get_kachery_dir import get_kachery_dir, use_sandbox  # noqa: F401
from .client.mutable_local import (  # noqa: F401
    delete_mutable_folder_local,
    delete_mutable_local,
    get_mutable_local,
    set_mutable_local,
)
from .client._sha1_of_dict import sha1_of_dict  # noqa: F401
from .client._custom_storage_backend import set_custom_storage_backend  # noqa: F401


# read the version from thisdir/version.txt
thisdir = os.path.dirname(os.path.realpath(__file__))
with open(os.path.join(thisdir, "version.txt")) as f:
    __version__ = f.read().strip()



import os
import sys
from typing import Union
from .load_file import load_file


# returns None if write_to_stdout is True
def load_bytes(
    uri: str, start: Union[int, None], end: Union[int, None], *, write_to_stdout=False
):
    local_path = load_file(uri)
    if local_path is None:
        return None

    return _load_bytes_from_local_file(
        local_path, start=start, end=end, write_to_stdout=write_to_stdout
    )


def _load_bytes_from_local_file(
    local_fname: str,
    *,
    start: Union[int, None] = None,
    end: Union[int, None] = None,
    write_to_stdout: bool = False
):
    size0 = os.path.getsize(local_fname)
    if start is None:
        start = 0
    if end is None:
        end = size0
    if start < 0 or start > size0 or end < start or end > size0:
        raise Exception(
            "Invalid start/end range for file of size {}: {} - {}".format(
                size0, start, end
            )
        )
    if start == end:
        return bytes()
    with open(local_fname, "rb") as f:
        f.seek(start)
        if write_to_stdout:
            ii = start
            while ii < end:
                nn = min(end - ii, 4096)
                data0 = f.read(nn)
                ii = ii + nn
                sys.stdout.buffer.write(data0)
            return None
        else:
            return f.read(end - start)



import os
import shutil
from typing import Union
from .get_kachery_dir import get_kachery_dir
from ._fs_operations import _makedirs


def set_mutable_local(key: str, value: str):
    _assert_valid_key(key)
    kachery_dir = get_kachery_dir()
    file_name = f"{kachery_dir}/mutables/{key}"
    parent_dir = os.path.dirname(file_name)
    if not os.path.exists(parent_dir):
        _makedirs(parent_dir)
    with open(file_name, "w") as f:
        f.write(value)
    # _chmod_file(file_name)


def _assert_valid_key(key: str):
    a = key.split("/")
    if len(a) < 1:
        raise Exception(f"Invalid mutable key: {key}")
    for i in range(len(a)):
        b = a[i]
        if i < len(a) - 1:
            if not b.startswith("@"):
                raise Exception(f"Invalid mutable key: {key}")
        else:
            if b.startswith("@"):
                raise Exception(f"Invalid mutable key: {key}")
        if b == "." or b == "..":
            raise Exception(f"Invalid mutable key: {key}")


def _assert_valid_folder_key(key: str):
    a = key.split("/")
    if len(a) < 1:
        raise Exception(f"Invalid mutable folder key: {key}")
    for i in range(len(a)):
        b = a[i]
        if not b.startswith("@"):
            raise Exception(f"Invalid mutable folder key: {key}")
        if b == "." or b == "..":
            raise Exception(f"Invalid mutable folder key: {key}")


def get_mutable_local(key: str, default_value: Union[None, str] = None):
    _assert_valid_key(key)
    kachery_dir = get_kachery_dir()
    file_name = f"{kachery_dir}/mutables/{key}"
    if not os.path.exists(file_name):
        return default_value
    with open(file_name, "r") as f:
        return f.read()


def delete_mutable_local(key: str):
    kachery_dir = get_kachery_dir()
    a = key.split("/")
    last_str = a[-1]
    if last_str.startswith("@"):
        raise Exception(f"Mutable key is a folder: {key}")
    else:
        _assert_valid_key(key)
        file_name = f"{kachery_dir}/mutables/{key}"
        if not os.path.exists(file_name):
            raise Exception(f"Mutable does not exist: {key}")
        os.unlink(file_name)


def delete_mutable_folder_local(key: str):
    kachery_dir = get_kachery_dir()
    a = key.split("/")
    last_str = a[-1]
    if last_str.startswith("@"):
        # deleting a folder
        _assert_valid_folder_key(key)
        folder_name = f"{kachery_dir}/mutables/{key}"
        if not os.path.exists(folder_name):
            raise Exception(f"Mutable folder does not exist: {key}")
        shutil.rmtree(folder_name)
    else:
        raise Exception(f"Not a mutable folder: {key}")


# def _add_snails(x: str):
#     # some/path/to/key -> @some/@path/@to/key
#     a = x.split('/')
#     for i in range(len(a) - 1):
#         a[i] = '@' + a[i]
#     return '/'.join(a)



from setuptools import setup, find_packages

# read version from kachery/version.txt
with open("kachery/version.txt") as f:
    __version__ = f.read().strip()

setup(
    name="kachery",
    version=__version__,
    author="Jeremy Magland, Luiz Tauffer, Alessio Buccino, Ben Dichter",
    author_email="jmagland@flatironinstitute.org",
    url="https://github.com/magland/kachery",
    description="",
    packages=find_packages(),
    include_package_data=True,
    package_data={"kachery": ["version.txt"]},
    install_requires=[
        "click",
        "simplejson",
        "numpy",
        "requests",
    ],
    entry_points={
        "console_scripts": [
            "kachery = kachery.cli:cli",
            "kachery-store = kachery.cli:store_file",
            "kachery-load = kachery.cli:load_file",
            "kachery-load-info = kachery.cli:load_file_info",
            "kachery-cat = kachery.cli:cat_file",
            "kachery-store-local = kachery.cli:store_file_local",
        ]
    },
)
